==> MuSE-0.7.2.tar.gz: MuSE-0.7.2/pipe.cpp <==
/*
  Copyright (c) 2001 Charles Samuels <charles@kde.org>
  Copyright (c) 2002 Denis Rojo <jaromil@dyne.org>
  
this pipe class was first written by Charles Samuels
and then heavily mutilated and optimized by Denis Rojo

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
  
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
   
You should have received a copy of the GNU Library General Public License
along with this library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

$Id:

*/

#include <pipe.h>
#include <iostream.h>
#include <stdlib.h>
#include <audioproc.h>
#include <jutils.h>
#include <generic.h>
#include <config.h>


#define MIN(a,b) (a<=b) ? a : b; 

/*
  start is a pointer to the first character that goes out
  end is a pointer to the last character to go out
*/

Pipe::Pipe(int size) {
  pipesize = size;
  buffer=malloc(pipesize);
  bufferEnd=(char*)buffer+size;
  end=start=buffer;
}

Pipe::~Pipe() {
  free(buffer);
}

int Pipe::read(int length, char fill, void *data) {
  int len=Pipe::read(length, data);
  if (len!=-1)
    ::memset((char*)data, fill, (char*)length-(char*)len);
  return len;
}

int Pipe::mix16stereo(int samples, int *mix) {
  int buffered = size();
  int c, cc;

  int length = samples<<2;
  int *pp = mix;

  //  if (!buffered) return -1;  

  if (buffered<length) {
    //    func("Pipe::mix16stereo(%i,%p) : only %i bytes in the pipe",
    //    samples,mix,buffered);
    return -1;
  }
  
  int origLength=length;
  
  while (buffered && length)	{
				
    /* |buffer*****|end-----------|start********|bufferEnd
       |buffer-----|start*********|end----------|bufferEnd */
    
    int len = MIN(length,buffered);
    
    int currentBlockSize = (char*)bufferEnd-(char*)start;
    currentBlockSize=MIN(currentBlockSize,len);

    /* mix */
    cc = currentBlockSize>>1;
    for(c=0; c<cc ;c++)
      pp[c] += (int) ((IN_DATATYPE*)start)[c];
    /* --- */

    (char*)start += currentBlockSize;
    len -= currentBlockSize;
    (char*)pp += currentBlockSize;
    length -= currentBlockSize;
    if ((end!=buffer) && (start==bufferEnd))
      start = buffer;
    
    if (len) { /* short circuit */
      
      /* mix */
      cc = len>>1;
      for(c=0; c<cc ;c++)
	pp[c] += (int) ((IN_DATATYPE*)start)[c];
      /* --- */

      (char*)pp += len;
      (char*)start += len;
      length -= len;
      if ((end!=buffer) && (start==bufferEnd))
	start = buffer;
    }
  }
  return origLength-length;
}  
	
int Pipe::read(int length, void *data) {
  int buffered = size();
  if (!buffered) return -1;
  if (buffered<length) {
    warning("Pipe::read(%i,%p) : can't read all that data from the pipe",length,data);
    return -1;
  }
  
  int origLength=length;
  
  while (buffered && length)	{
				
    /* |buffer*****|end-----------|start********|bufferEnd
       |buffer-----|start*********|end----------|bufferEnd */
    
    int len = MIN(length,buffered);
    
    int currentBlockSize = (char*)bufferEnd-(char*)start;
    currentBlockSize=MIN(currentBlockSize,len);
    memcpy(data, start, currentBlockSize);
    
    (char*)start += currentBlockSize;
    len -= currentBlockSize;
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((end!=buffer) && (start==bufferEnd))
      start = buffer;
    
    if (len) { /* short circuit */
      memcpy(data, start, len);
      (char*)data += len;
      (char*)start += len;
      length -= len;
      if ((end!=buffer) && (start==bufferEnd))
	start = buffer;
    }
  }	
  return origLength-length;
}

int Pipe::write(int length, void *data) {
  while (length) {
    
    /* |buffer-----|end***********|start--------|bufferEnd
       |buffer*****|start---------|end**********|bufferEnd */
		
    int len=MIN(length, space()-1);
    
    int currentBlockSize = (char*)bufferEnd-(char*)end;
    currentBlockSize=MIN(currentBlockSize, len);
    ::memcpy(end, data, currentBlockSize);
    
    (char*)end += currentBlockSize;
    
    len -= currentBlockSize;
    
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((start!=buffer) && (end==bufferEnd))
      end = buffer;
		
    if (len) { // short circuit		
      ::memcpy(end, data, len);
      (char*)data += len;
      (char*)end += len;
      length -= len;
      
      if ((start!=buffer) && (end==bufferEnd))
	end = buffer;
    }
  }
  return length;
}

// |buffer******|end--------------|start**************|bufferEnd
// |buffer-----|start**************|end---------------|bufferEnd
int Pipe::size() const {
  int val;
  if ((char*)end >= (char*)start)
    val = (char*)end-(char*)start;
  else 
    val = ((char*)bufferEnd-(char*)start)+((char*)end-(char*)buffer);
  return val;
}

// |buffer------|end**************|start--------------|bufferEnd
// |buffer*****|start--------------|end***************|bufferEnd
int Pipe::space() const {
  return ((char*)bufferEnd-(char*)buffer)-size();
}

void Pipe::flush() {
  /*
  int i = size();
  if(i>0) flush(size());
  */
  bufferEnd=(char*)buffer+pipesize;
  end=start=buffer;
}

void Pipe::flush(int bytes) {
  void *temp = malloc(bytes);
  read(bytes, temp);
  free(temp);
}

==> MuSE-0.8.tar.gz: MuSE-0.8/pipe.cpp <==
/*
  Copyright (c) 2001 Charles Samuels <charles@kde.org>
  Copyright (c) 2002 Denis Rojo <jaromil@dyne.org>
  
this pipe class was first written by Charles Samuels
and then heavily mutilated and optimized by Denis Rojo

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
  
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
   
You should have received a copy of the GNU Library General Public License
along with this library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

"$Id: pipe.cpp,v 1.16 2002/12/30 08:47:27 jaromil Exp $"

*/

#include <pipe.h>
#include <iostream.h>
#include <stdlib.h>
#include <audioproc.h>
#include <jutils.h>
#include <generic.h>
#include <config.h>


#define MIN(a,b) (a<=b) ? a : b; 

#define _SIZE(val) \
  int val; \
  if ((char*)end >= (char*)start) \
    val = (char*)end-(char*)start; \
  else  \
    val = ((char*)bufferEnd-(char*)start)+((char*)end-(char*)buffer);

#define _SPACE(val) \
  _SIZE(__size); \
  val = ((char*)bufferEnd-(char*)buffer)-__size;

/*
  start is a pointer to the first character that goes out
  end is a pointer to the last character to go out
*/

Pipe::Pipe(int size) {
  pipesize = size;
  buffer=malloc(pipesize);
  bufferEnd=(char*)buffer+size;
  end=start=buffer;
  _thread_init();
  unlock();
}

Pipe::~Pipe() {
  unlock();
  free(buffer);
  _thread_destroy();
}

int Pipe::read(int length, char fill, void *data) {
  int len=Pipe::read(length, data);
  lock();
  if (len!=-1)
    ::memset((char*)data, fill, (char*)length-(char*)len);

  unlock();
  return len;
}

int Pipe::mix16stereo(int samples, int32_t *mix) {
  lock();
  /* int buffered = size(); */
  _SIZE(buffered);
  int c, cc;

  int length = samples<<2;
  int32_t *pp = mix;

  //  if (!buffered) return -1;  

  if (buffered<length) {
    //    func("Pipe::mix16stereo(%i,%p) : only %i bytes in the pipe",
    //    samples,mix,buffered);
    unlock();
    return -1;
  }
  
  int origLength=length;
  
  while (buffered && length)	{
				
    /* |buffer*****|end-----------|start********|bufferEnd
       |buffer-----|start*********|end----------|bufferEnd */
    
    int len = MIN(length,buffered);
    
    int currentBlockSize = (char*)bufferEnd-(char*)start;
    currentBlockSize=MIN(currentBlockSize,len);

    /* mix */
    cc = currentBlockSize>>1;
    for(c=0; c<cc ;c++)
      pp[c] += (int32_t) ((IN_DATATYPE*)start)[c];
    /* --- */

    (char*)start += currentBlockSize;
    len -= currentBlockSize;
    (char*)pp += currentBlockSize;
    length -= currentBlockSize;
    if ((end!=buffer) && (start==bufferEnd))
      start = buffer;
    
    if (len) { /* short circuit */
      
      /* mix */
      cc = len>>1;
      for(c=0; c<cc ;c++)
	pp[c] += (int) ((IN_DATATYPE*)start)[c];
      /* --- */

      (char*)pp += len;
      (char*)start += len;
      length -= len;
      if ((end!=buffer) && (start==bufferEnd))
	start = buffer;
    }
  }

  unlock();
  return origLength-length;
}  
	
int Pipe::read(int length, void *data) {
  lock();

  /*  int buffered = size(); */
  _SIZE(buffered);
  if (buffered<length) {
    //    warning("Pipe::read(%i,%p) : not enough data in the pipe",length,data);
    unlock();
    return -1;
  }
  
  int origLength=length;
  
  while (buffered && length)	{
				
    /* |buffer*****|end-----------|start********|bufferEnd
       |buffer-----|start*********|end----------|bufferEnd */
    
    int len = MIN(length,buffered);
    
    int currentBlockSize = (char*)bufferEnd-(char*)start;
    currentBlockSize=MIN(currentBlockSize,len);
    memcpy(data, start, currentBlockSize);
    
    (char*)start += currentBlockSize;
    len -= currentBlockSize;
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((end!=buffer) && (start==bufferEnd))
      start = buffer;
    
    if (len) { /* short circuit */
      memcpy(data, start, len);
      (char*)data += len;
      (char*)start += len;
      length -= len;
      if ((end!=buffer) && (start==bufferEnd))
	start = buffer;
    }
  }
  
  unlock();
  return origLength-length;
}

int Pipe::write(int length, void *data) {
  int spc;
  lock();

  _SPACE(spc);
  if (length>(spc-1)) {
    //      warning("Pipe::write(%i,%p) : not enough space in the pipe",length,data);
    unlock();
    return -1;
  }

  while (length) {
    
    /* |buffer-----|end***********|start--------|bufferEnd
       |buffer*****|start---------|end**********|bufferEnd */
    _SPACE(spc);
    int len=MIN(length, spc-1);
    
    int currentBlockSize = (char*)bufferEnd-(char*)end;
    currentBlockSize=MIN(currentBlockSize, len);
    ::memcpy(end, data, currentBlockSize);
    
    (char*)end += currentBlockSize;
    
    len -= currentBlockSize;
    
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((start!=buffer) && (end==bufferEnd))
      end = buffer;
		
    if (len) { // short circuit		
      ::memcpy(end, data, len);
      (char*)data += len;
      (char*)end += len;
      length -= len;
      
      if ((start!=buffer) && (end==bufferEnd))
	end = buffer;
    }
  }
  unlock();
  return length;
}

// |buffer******|end--------------|start**************|bufferEnd
// |buffer-----|start**************|end---------------|bufferEnd
int Pipe::size() {
  lock();
  _SIZE(res);
  unlock();
  return res;
}

// |buffer------|end**************|start--------------|bufferEnd
// |buffer*****|start--------------|end***************|bufferEnd
int Pipe::space() {
  lock();
  int res; _SPACE(res);
  unlock();
  return res;
}

void Pipe::flush() {
  /*
  int i = size();
  if(i>0) flush(size());
  */
  lock();
  bufferEnd=(char*)buffer+pipesize;
  end=start=buffer;
  unlock();
}

void Pipe::flush(int bytes) {
  lock();
  void *temp = malloc(bytes);
  read(bytes, temp);
  free(temp);
  unlock();
}
